/* kas_isofs.c - A shared library to handle MacSCII <-> ISO 8859-1
                 character set conversions for filename handling in
	         Xinets K/A-Share server products. It's a hack
		 and it would have been much better to have this
		 functionality directly in K/A-Share...

   Copyright (c) 1996-1998 Peter Eriksson <pen@lysator.liu.se>

   Version 1.4
   

   This shared library module is free software; you can redistribute it
   and/or modify it any way you like - as long as you write down all
   the modifications to it in this file, along with your name and
   email address. 
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

   
   You may want to uncomment/add more filename suffixes in the
   array "isotxtsuffix" below. By default no file content
   conversions will be done. Enable by the "-DCONTENTCONV" compilation
   flag. You may also want to edit the /var/adm/appletalk/filetype
   file so that there is a line like:

   	0       suffix  txt     :TEXT:uxtt:     # Generic text file

   K/A-Share will think that text files containing ISO 8859-1
   characters are binary files if you don't do that...

   
   Compile with:

       gcc -fPIC -O -c kas_isofs.c ; ld -G -o kas_isofs.so kas_isofs.o

   Install the "kas_isofs.so" file into /usr/etc/appletalk.
   
   Then rename /usr/etc/appletalk/ksd to "ksd.bin"
   and install a shellscript in it's place that looks like
   this:

   	#!/bin/sh
 
	LD_PRELOAD=/usr/etc/appletalk/kas_isofs.so
	export LD_PRELOAD
	exec /usr/etc/appletalk/ksd.bin $*

	
   TODO:

   	If a syscall fails with ENOENT, should perhaps try to
	rename the file from the ":XX" to the ISO formats,
	and then retry the operation instead of opening the
	file with the old name. But we must also handle
	renaming the directory pathname components at the
	same time...
       
   
   This library redefines the following syscalls:

   open()
   unlink()
   rename()
   stat()
   lstat()
   opendir()
   readdir()
   readdir_r()
   fopen()
   access()
   chdir()
   chmod()
   chown()
   creat()
   lchown()
   mkdir()
   readlink()
   rmdir()
   symlink()
*/

#include <stdio.h>
#include <alloca.h>

#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <errno.h>

#include <sys/time.h>
#include <sys/resource.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdarg.h>

#include <dirent.h>

#define PATH_MAX 2048

int *flags = NULL;


#define FLG_ISO     1

/*
** Filenames ending with suffixes in the list below will
** get their contents converted to/from Macscii automatically
*/
static char *isotxtsuffix[] =
{
#ifdef CONTENTCONV
    ".txt",
    ".html",
    ".shtml",
#endif
    NULL
};


static void setflag(int fd, const char *path)
{
    struct rlimit rlb;
    int max_fds = 2048;
    int len, i;

    
    
    if (flags == NULL)
	flags = calloc(sizeof(int), max_fds);

    flags[fd] = 0;

    len = strlen(path);

    
    for (i = 0; isotxtsuffix[i] &&
	     strcmp(isotxtsuffix[i], path+len-strlen(isotxtsuffix[i])); i++)
	;

    if (isotxtsuffix[i])
	flags[fd] |= FLG_ISO;
}


/* Conversion table from `mac' charset to `latin1' charset.
   Generated mechanically by GNU recode 3.4.

   The recoding should be reversible.
*/

static unsigned char const mac_to_latin1[256] =
  {
      0,   1,   2,   3,   4,   5,   6,   7,	/*   0 -   7 */
      8,   9,  10,  11,  12,  13,  14,  15,	/*   8 -  15 */
     16,  17,  18,  19,  20,  21,  22,  23,	/*  16 -  23 */
     24,  25,  26,  27,  28,  29,  30,  31,	/*  24 -  31 */
     32,  33,  34,  35,  36,  37,  38,  39,	/*  32 -  39 */
     40,  41,  42,  43,  44,  45,  46,  47,	/*  40 -  47 */
     48,  49,  50,  51,  52,  53,  54,  55,	/*  48 -  55 */
     56,  57,  58,  59,  60,  61,  62,  63,	/*  56 -  63 */
     64,  65,  66,  67,  68,  69,  70,  71,	/*  64 -  71 */
     72,  73,  74,  75,  76,  77,  78,  79,	/*  72 -  79 */
     80,  81,  82,  83,  84,  85,  86,  87,	/*  80 -  87 */
     88,  89,  90,  91,  92,  93,  94,  95,	/*  88 -  95 */
     96,  97,  98,  99, 100, 101, 102, 103,	/*  96 - 103 */
    104, 105, 106, 107, 108, 109, 110, 111,	/* 104 - 111 */
    112, 113, 114, 115, 116, 117, 118, 119,	/* 112 - 119 */
    120, 121, 122, 123, 124, 125, 126, 127,	/* 120 - 127 */
    196, 197, 199, 201, 209, 214, 220, 225,	/* 128 - 135 */
    224, 226, 228, 227, 229, 231, 233, 232,	/* 136 - 143 */
    234, 235, 237, 236, 238, 239, 241, 243,	/* 144 - 151 */
    242, 244, 246, 245, 250, 249, 251, 252,	/* 152 - 159 */
    190, 176, 162, 163, 167, 130, 182, 223,	/* 160 - 167 */
    174, 169, 142, 180, 168, 173, 198, 216,	/* 168 - 175 */
    141, 177, 178, 179, 165, 181, 166, 135,	/* 176 - 183 */
    159, 185, 188, 170, 186, 189, 230, 248,	/* 184 - 191 */
    191, 161, 172, 146, 128, 129, 140, 171,	/* 192 - 199 */
    187, 131, 160, 192, 195, 213, 145, 147,	/* 200 - 207 */
    208, 132, 150, 148, 149, 144, 247, 215,	/* 208 - 215 */
    255, 153, 152, 164, 134, 221, 222, 151,	/* 216 - 223 */
    136, 183, 137, 139, 138, 194, 202, 193,	/* 224 - 231 */
    203, 200, 205, 206, 207, 204, 211, 212,	/* 232 - 239 */
    240, 210, 218, 219, 217, 155, 154, 133,	/* 240 - 247 */
    175, 157, 156, 158, 184, 253, 254, 143,	/* 248 - 255 */
  };

/* Conversion table from `latin1' charset to `mac' charset.
   Generated mechanically by GNU recode 3.4.

   The recoding should be reversible.
*/

static unsigned char const latin1_to_mac[256] =
  {
      0,   1,   2,   3,   4,   5,   6,   7,	/*   0 -   7 */
      8,   9,  10,  11,  12,  13,  14,  15,	/*   8 -  15 */
     16,  17,  18,  19,  20,  21,  22,  23,	/*  16 -  23 */
     24,  25,  26,  27,  28,  29,  30,  31,	/*  24 -  31 */
     32,  33,  34,  35,  36,  37,  38,  39,	/*  32 -  39 */
     40,  41,  42,  43,  44,  45,  46,  47,	/*  40 -  47 */
     48,  49,  50,  51,  52,  53,  54,  55,	/*  48 -  55 */
     56,  57,  58,  59,  60,  61,  62,  63,	/*  56 -  63 */
     64,  65,  66,  67,  68,  69,  70,  71,	/*  64 -  71 */
     72,  73,  74,  75,  76,  77,  78,  79,	/*  72 -  79 */
     80,  81,  82,  83,  84,  85,  86,  87,	/*  80 -  87 */
     88,  89,  90,  91,  92,  93,  94,  95,	/*  88 -  95 */
     96,  97,  98,  99, 100, 101, 102, 103,	/*  96 - 103 */
    104, 105, 106, 107, 108, 109, 110, 111,	/* 104 - 111 */
    112, 113, 114, 115, 116, 117, 118, 119,	/* 112 - 119 */
    120, 121, 122, 123, 124, 125, 126, 127,	/* 120 - 127 */
    196, 197, 165, 201, 209, 247, 220, 183,	/* 128 - 135 */
    224, 226, 228, 227, 198, 176, 170, 255,	/* 136 - 143 */
    213, 206, 195, 207, 211, 212, 210, 223,	/* 144 - 151 */
    218, 217, 246, 245, 250, 249, 251, 184,	/* 152 - 159 */
    202, 193, 162, 163, 219, 180, 182, 164,	/* 160 - 167 */
    172, 169, 187, 199, 194, 173, 168, 248,	/* 168 - 175 */
    161, 177, 178, 179, 171, 181, 166, 225,	/* 176 - 183 */
    252, 185, 188, 200, 186, 189, 160, 192,	/* 184 - 191 */
    203, 231, 229, 204, 128, 129, 174, 130,	/* 192 - 199 */
    233, 131, 230, 232, 237, 234, 235, 236,	/* 200 - 207 */
    208, 132, 241, 238, 239, 205, 133, 215,	/* 208 - 215 */
    175, 244, 242, 243, 134, 221, 222, 167,	/* 216 - 223 */
    136, 135, 137, 139, 138, 140, 190, 141,	/* 224 - 231 */
    143, 142, 144, 145, 147, 146, 148, 149,	/* 232 - 239 */
    240, 150, 152, 151, 153, 155, 154, 214,	/* 240 - 247 */
    191, 157, 156, 158, 159, 253, 254, 216,	/* 248 - 255 */
  };

static int (*next_open)();
static int (*next_unlink)();
static int (*next_rename)();
static int (*next_stat)();
static int (*next_lstat)();
static int (*next_access)();
static int (*next_chdir)();
static int (*next_chmod)();
static int (*next_chown)();
static int (*next_creat)();
static int (*next_lchown)();
static int (*next_mkdir)();
static int (*next_readlink)();
static int (*next_rename)();
static int (*next_rmdir)();
static int (*next_symlink)();

static DIR *(*next_opendir)();
static struct dirent *(*next_readdir_r)();

static FILE *(*next_fopen)();

static int (*next_read)();
static int (*next_write)();



static int real_rename(const char *path1, const char *path2)
{
    if (next_rename == NULL)
    {
	next_rename = (int (*)()) dlsym(RTLD_NEXT, "rename");
	if (next_rename == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    return next_rename(path1, path2);
}    


static int hex2num(int c)
{
    if (isdigit(c))
	return c - '0';
    else if (c >= 'A' && c <= 'F')
	return c - 'A' + 10;
    else if (c >= 'a' && c <= 'f')
	return c - 'a' + 10;
    else
	return -1;
}


/* Convert ":XX" -> ISO characters.
**
*/
static void kas2iso(const char *in, char *out, int size)
{
    char *end = out + size - 4;
    
    while (*in && out < end)
    {
	if (in[0] == ':' && in[1] == ':')
	{
	    *out++ = *in++;
	    *out++ = *in++;
	}
	else if (in[0] == ':' && in[1] && in[2] &&
	    (isxdigit(in[1]) && !islower(in[1]) &&
	     isxdigit(in[2]) && !islower(in[2])))
	{
	    int code;

	    
	    code = mac_to_latin1[hex2num(in[1])*16+hex2num(in[2])];

	    if (code == '/')
	    {
		*out++ = ':';
		*out++ = ':';
	    }
	    else
		*out++ = code;

	    in += 3;
	}
	else
	    *out++ = *in++;
    }

    *out = '\0';
}


static void iso2kas(const unsigned char *in, char *out, int size)
{
    char *end = out + size - 4;
    
    while (*in && out < end)
    {
	if (in[0] == ':' && in[1] == ':')
	{
	    sprintf(out, ":%02X", '/');
	    out += 3;
	    in += 2;
	}
	else if (in[0] == ':' && isxdigit(in[1]) && isxdigit(in[2]))
	{
	    *out++ = *in++;
	    *out++ = *in++;
	    *out++ = *in++;
	}	
	else if (/* *in < ' ' || */ *in >= 128 || *in == ':')
	{
	    int code = latin1_to_mac[*in];
	    sprintf(out, ":%02X", code);
	    out += 3;
	    in++;
	}
	else
	    *out++ = *in++;
    }

    *out = '\0';
}


int open(const char *path, int oflag, ...)
{
    int code;
    char pathbuf[PATH_MAX];
    


    if (next_open == NULL)
    {
	next_open = (int (*)()) dlsym(RTLD_NEXT, "open");
	if (next_open == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    if (oflag & O_CREAT)
    {
	mode_t mode;
	va_list ap;

	
	va_start(ap, oflag);

	mode = va_arg(ap, mode_t);
	code = next_open(pathbuf, oflag, mode);
	if (code < 0 && errno == ENOENT)
	    code = next_open(path, oflag, mode);

	va_end(ap);
    }
    else
    {
	code = next_open(pathbuf, oflag);
	if (code < 0 && errno == ENOENT)
	    code = next_open(path, oflag);
    }

    if (code >= 0)
	setflag(code, path);
    
    return code;
}




int unlink(const char *path)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_unlink == NULL)
    {
	next_unlink = (int (*)()) dlsym(RTLD_NEXT, "unlink");
	if (next_unlink == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_unlink(pathbuf);
    if (code < 0 && errno == ENOENT)
	code = next_unlink(path);

    return code;
}



int stat(const char *path, struct stat *buf)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_stat == NULL)
    {
	next_stat = (int (*)()) dlsym(RTLD_NEXT, "stat");
	if (next_stat == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_stat(pathbuf, buf);
    if (code < 0 && errno == ENOENT)
	code = next_stat(path, buf);

    return code;
}

int lstat(const char *path, struct stat *buf)
{
    int code;
    char pathbuf[PATH_MAX];


    if (next_lstat == NULL)
    {
	next_lstat = (int (*)()) dlsym(RTLD_NEXT, "lstat");
	if (next_lstat == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_lstat(pathbuf, buf);
    if (code < 0 && errno == ENOENT)
	code = next_lstat(path, buf);

    return code;
}


DIR *opendir(const char *path)
{
    DIR *dp;
    char pathbuf[PATH_MAX];
    

    if (next_opendir == NULL)
    {
	next_opendir = (DIR *(*)()) dlsym(RTLD_NEXT, "opendir");
	if (next_opendir == NULL)
	{
	    errno = EINVAL;
	    return NULL;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    dp = next_opendir(pathbuf);

    return dp;
}


struct dirent *readdir_r(DIR *dp, char *result)
{
    struct dirent *dep;
    char pathbuf[PATH_MAX];

    
    if (next_readdir_r == NULL)
    {
	next_readdir_r = (struct dirent*(*)()) dlsym(RTLD_NEXT, "readdir_r");
	if (next_readdir_r == NULL)
	{
	    errno = EINVAL;
	    return NULL;
	}
    }

  Again:
    dep = next_readdir_r(dp, result);
    if (dep)
    {
#if 0
	/* Sanity check */
	if (strlen(dep->d_name) > 31)
	    goto Again;
#endif
	
	iso2kas(dep->d_name, pathbuf, sizeof(pathbuf));
	strcpy(dep->d_name, pathbuf);
	dep->d_reclen = sizeof(*dep)+strlen(dep->d_name);
    }
    
    return dep;
}



struct dirent *readdir(DIR *dp)
{
    struct dirent *dep;
    char resbuf[PATH_MAX+128];
    char pathbuf[PATH_MAX];


    return readdir_r(dp, resbuf);
}



FILE *fopen(const char *path, const char *mode)
{
    FILE *fp;
    char pathbuf[PATH_MAX];
    

    if (next_fopen == NULL)
    {
	next_fopen = (FILE *(*)()) dlsym(RTLD_NEXT, "fopen");
	if (next_fopen == NULL)
	{
	    errno = EINVAL;
	    return NULL;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    fp = next_fopen(pathbuf, mode);
    if (fp == NULL && errno == ENOENT)
	fp = next_fopen(path, mode);

    return fp;
}



int access(const char *path, int amode)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_access == NULL)
    {
	next_access = (int (*)()) dlsym(RTLD_NEXT, "access");
	if (next_access == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_access(pathbuf, amode);
    if (code < 0 && errno == ENOENT)
	code = next_access(path, amode);

    return code;
}

int creat(const char *path, mode_t mode)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_creat == NULL)
    {
	next_creat = (int (*)()) dlsym(RTLD_NEXT, "creat");
	if (next_creat == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_creat(pathbuf, mode);
    if (code < 0 && errno == ENOENT)
	code = next_creat(path, mode);

    if (code >= 0)
	setflag(code, path);
    
    return code;
}

int chdir(const char *path)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_chdir == NULL)
    {
	next_chdir = (int (*)()) dlsym(RTLD_NEXT, "chdir");
	if (next_chdir == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_chdir(pathbuf);
    if (code < 0 && errno == ENOENT)
	code = next_chdir(path);

    return code;
}


int chmod(const char *path, mode_t mode)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_chmod == NULL)
    {
	next_chmod = (int (*)()) dlsym(RTLD_NEXT, "chmod");
	if (next_chmod == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_chmod(pathbuf, mode);
    if (code < 0 && errno == ENOENT)
	code = next_chmod(path, mode);

    return code;
}


int chown(const char *path, uid_t owner, gid_t group)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_chown == NULL)
    {
	next_chown = (int (*)()) dlsym(RTLD_NEXT, "chown");
	if (next_chown == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_chown(pathbuf, owner, group);
    if (code < 0 && errno == ENOENT)
	code = next_chown(path, owner, group);

    return code;
}


int lchown(const char *path, uid_t owner, gid_t group)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_lchown == NULL)
    {
	next_lchown = (int (*)()) dlsym(RTLD_NEXT, "lchown");
	if (next_lchown == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_lchown(pathbuf, owner, group);
    if (code < 0 && errno == ENOENT)
	code = next_lchown(path, owner, group);

    return code;
}

int mkdir(const char *path, mode_t mode)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_mkdir == NULL)
    {
	next_mkdir = (int (*)()) dlsym(RTLD_NEXT, "mkdir");
	if (next_mkdir == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_mkdir(pathbuf, mode);
    if (code < 0 && errno == ENOENT)
	code = next_mkdir(path, mode);

    return code;
}


int readlink(const char *path, void *buf, size_t bufsiz)
{
    int code;
    char pathbuf1[PATH_MAX];
    char pathbuf2[PATH_MAX];
    

    if (next_readlink == NULL)
    {
	next_readlink = (int (*)()) dlsym(RTLD_NEXT, "readlink");
	if (next_readlink == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf1, sizeof(pathbuf1));
    
    code = next_readlink(pathbuf1, pathbuf2, sizeof(pathbuf2));
    if (code < 0 && errno == ENOENT)
	code = next_readlink(path, pathbuf2, sizeof(pathbuf2));

    if (code > 0)
    {
	pathbuf2[code] = '\0';
	iso2kas(pathbuf2, pathbuf1, sizeof(pathbuf1));
	code = strlen(pathbuf1);
	if (code > bufsiz)
	    code = bufsiz;

	strncpy(buf, pathbuf1, code);
    }

    return code;
}




int rmdir(const char *path)
{
    int code;
    char pathbuf[PATH_MAX];
    

    if (next_rmdir == NULL)
    {
	next_rmdir = (int (*)()) dlsym(RTLD_NEXT, "rmdir");
	if (next_rmdir == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path, pathbuf, sizeof(pathbuf));
    
    code = next_rmdir(pathbuf);
    if (code < 0 && errno == ENOENT)
	code = next_rmdir(path);

    return code;
}


int rename(const char *path1, const char *path2)
{
    int code;
    char pathbuf1[PATH_MAX], pathbuf2[PATH_MAX];
    

    kas2iso(path1, pathbuf1, sizeof(pathbuf1));
    kas2iso(path2, pathbuf2, sizeof(pathbuf2));
    
    code = real_rename(pathbuf1, pathbuf2);
    if (code < 0 && errno == ENOENT)
	code = real_rename(path1, pathbuf2);

    return code;
}


int symlink(const char *path1, const char *path2)
{
    int code;
    char pathbuf1[PATH_MAX], pathbuf2[PATH_MAX];
    

    if (next_symlink == NULL)
    {
	next_symlink = (int (*)()) dlsym(RTLD_NEXT, "symlink");
	if (next_symlink == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    kas2iso(path1, pathbuf1, sizeof(pathbuf1));
    kas2iso(path2, pathbuf2, sizeof(pathbuf2));
    
    code = next_symlink(pathbuf1, pathbuf2);
    if (code < 0 && errno == ENOENT)
	code = next_symlink(pathbuf1, path2);

    return code;
}



int read(int fd, unsigned char *buf, size_t len)
{
    int code;
    

    if (next_read == NULL)
    {
	next_read = (int (*)()) dlsym(RTLD_NEXT, "read");
	if (next_read == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    code = next_read(fd, buf, len);

    if (flags[fd] & FLG_ISO)
    {
	int i;

	
	for (i = 0; i < code; i++)
	    buf[i] = latin1_to_mac[buf[i]];
    }
    
    return code;
}



int write(int fd, unsigned char *buf, size_t len)
{
    int code;
    char *tbuf;

    
    if (next_write == NULL)
    {
	next_write = (int (*)()) dlsym(RTLD_NEXT, "write");
	if (next_write == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}
    }

    if (flags[fd] & FLG_ISO)
    {
	int i;

	
	tbuf = alloca(len);
	if (tbuf == NULL)
	{
	    errno = EINVAL;
	    return -1;
	}

	for (i = 0; i < len; i++)
	    tbuf[i] = mac_to_latin1[buf[i]];
    }
    else
	tbuf = buf;
    
    code = next_write(fd, tbuf, len);

    return code;
}




#ifdef WITHMAIN
main(int argc, char *argv[])
{
    char buf[PATH_MAX];

    kas2iso(argv[1], buf, sizeof(buf));
    puts(buf);

    iso2kas(argv[1], buf, sizeof(buf));
    puts(buf);
exit(0);
}
#endif



